"""
A python gateway to the Saxon 9 XSLT transform library.

We haven't been able to find a useful native XSLT 2.0 processor in python.
XSLTGateway uses the py4j library to connect to the Saxon-HE processor. The
package supports caching/pre-compilation of xslt transforms, and includes a
built in transformation that implements the OMG XML to JSON specification.

Links:
    py4j: https://www.py4j.org/
    Saxon-HD: http://saxon.sourceforge.net/
    Spec: http://www.omg.org/spec/CTS2/1.2/PDF/ (Appendix B)
    Transformation: http://informatics.mayo.edu/cts2/services/xmlToJson/

Requirements:
    java 1.7


A tiny example:
    Command line:
    > cd XSLTGateway/javagateway
    > ./run.sh
        <path>/XSLTGateway/javagateway
        Gateway Server Started on default port

    > python
    >>> from XSLTGateway import Gateway

    >>> gw = Gateway()
    Starting Java gateway on port: 2533

    >>> print(gw.to_json("<doc name='foo'><entry>Page 1</entry></doc>"))
    {
      "doc": {
        "name": "foo",
        "entry": "Page 1"
      }
    }

    >>> gw.add_transform("test", "examples/test.xsl")
    >>> print(gw.transform("test", "<doc name='foo'><entry>Page 1</entry></doc>"))
    <html>
       <head>
          <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
          <title>XSLT Output</title>
       </head>
       <body>
          <h1>foo</h1>
          <p>Page 1</p>
       </body>
    </html>

"""
__docformat__ = "restructuredtext en"

# The format of the __version__ line is matched by a regex in setup.py
__version__ = "1.0.0-rc.1"
__date__ = "2015/10/29"

__all__ = [
    'XSLTGateway',
    'XSLTLibrary',
]

import sys
# This could be easily corrected if needed
assert sys.version_info >= (3, 1, 0), "XSLTGateway requires Python 3.1 or higher"
del sys

import logging
_LOGGER = logging.getLogger("XSLTGateway")
_LOGGER.info("XSLTGateway Version: %s" % __version__)

from XSLTGateway.XSLTGateway import Gateway



# -*- coding: utf-8 -*-
# Copyright (c) 2014, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the <ORGANIZATION> nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

import os


class XSLTLibrary(dict):
    """ A library of XSLT modules -- either as files or as xslt text.
    """
    @staticmethod
    def is_xslt(text):
        return text.startswith('<?xml ') and text.strip().endswith('</xsl:stylesheet>')

    @staticmethod
    def is_xslt_file(fname):
        with open(fname) as infile:
            return XSLTLibrary.is_xslt(infile.read())

    def __setitem__(self, key, value):
        """
        Add a file or block of text to the dictionary
        @param key: value key
        @param value: either a file or a block of xslt
        """
        if self.is_xslt(value) or os.path.isfile(value):
            super(XSLTLibrary, self).__setitem__(key, value)
        else:
            raise ValueError("'%s' is neither XSLT nor the name of a file" % key)

    def __getitem__(self, key):
        val = super(XSLTLibrary, self).__getitem__(key)
        if self.is_xslt(val):
            return val
        with open(val) as f:
            return f.read()



# -*- coding: utf-8 -*-
# Copyright (c) 2014, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the <ORGANIZATION> nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import os

import pyjxslt

testXSLT = """<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    exclude-result-prefixes="xs xd"
    version="2.0">
</xsl:stylesheet>
"""

testNonXSLT = """<?xml ... >"""


class TestLibrary(unittest.TestCase):
    def testText(self):
        lib = pyjxslt.XSLTLibrary()
        lib['k1'] = testXSLT
        self.assertEqual(lib['k1'], testXSLT)

    def testNonXSLT(self):
        lib = pyjxslt.XSLTLibrary()
        self.assertRaises(ValueError, lib.__setitem__, *('k2', testNonXSLT))

    def testFile(self):
        lib = pyjxslt.XSLTLibrary()
        lib['k3'] = os.path.join('data', 'file1.xslt')
        with open(os.path.join('data', 'file1.xslt')) as f:
            self.assertEqual(f.read(), lib['k3'])

    def testNoFile(self):
        lib = pyjxslt.XSLTLibrary()
        self.assertRaises(ValueError, lib.__setitem__, *('k4', 'nofile.xslt'))

    def testNonXSLTFile(self):
        lib = pyjxslt.XSLTLibrary()
        self.assertRaises(ValueError, lib.__setitem__, *('k4', 'file2.txt'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the <ORGANIZATION> nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import json
import socket
from functools import reduce

from py4j.java_gateway import JavaGateway, GatewayClient, Py4JNetworkError

from .XSLTLibrary import XSLTLibrary

DEFAULT_PORT = 25333
XML_TO_JSON_KEY = 'A69F49B0-2C34-4762-A27C-081B5FD4FF35'        # A safe key for the XML to JSON XSLT transformation


class Gateway(object):
    def __init__(self, port=DEFAULT_PORT, **_):
        """ Construct a new XSLT gateway.  This uses the py4j gateway to connect to a java server.

        @param port: py4j gateway port (default: DEFAULT_PORT)
        """
        self._gwPort = int(port)
        self._converters = {}
        self._xsltLibrary = XSLTLibrary()
        self._xsltFactory = None
        self._gateway = None
        self._add_json_xslt()
        self.reconnect()

    def reconnect(self):
        """ (Re)establish the gateway connection
        @return: True if connection was established
        """
        self._converters.clear()
        self._gateway = None
        self._xsltFactory = None
        try:
            # print("Starting Java gateway on port: %s" % self._gwPort)
            self._gateway = JavaGateway(GatewayClient(port=self._gwPort))
            self._xsltFactory = self._gateway.jvm.org.pyjxslt.XSLTTransformerFactory('')
            self._refresh_converters()
        except (socket.error, Py4JNetworkError) as e:
            print(e)
            self._gateway = None
            return False
        return True
    
    def gateway_connected(self, reconnect=True):
        """ Determine whether the gateway is connected
        @param reconnect: True means try to reconnect if not connected
        @return: True if the gateway is active
        """
        return self._gateway is not None or (reconnect and self.reconnect())

    def to_json(self, xml):
        ugly_json = self.transform(XML_TO_JSON_KEY, xml)
        try:
            rval = json.dumps(json.loads(ugly_json), indent=4)
        except json.JSONDecodeError:
            rval = ugly_json
        return rval

    def add_transform(self, key, xslt):
        """ Add or update a transform.

        @param key: Transform key to use when executing transformations
        @param xslt: Text or file name of an xslt transform

        """
        self._remove_converter(key)
        self._xsltLibrary[key] = xslt
        self._add_converter(key)

    def drop_transform(self, key):
        self._remove_converter(key)

    def _add_json_xslt(self):
        self.add_transform(XML_TO_JSON_KEY,
                           os.path.join(os.path.join(os.getcwd(), os.path.dirname(__file__)), 'xsl', 'XMLToJson.xsl'))

    def _refresh_converters(self):
        """ Refresh all of the converters in the py4j library
        @return: True if all converters were succesfully updated
        """
        self._converters.clear()
        return reduce(lambda a, b: a and b, [self._add_converter(k) for k in list(self._xsltLibrary.keys())], True)

    def _add_converter(self, key):
        # Do the checkConnected first, as, if the connection is isn't reestablished not much we can do
        if self.gateway_connected(reconnect=False) and key not in self._converters:
            try:
                self._converters[key] = self._xsltFactory.transformer(key, self._xsltLibrary[key])
                return True
            except socket.error as e:
                print(e)
                self._gateway = None
        return False

    def _remove_converter(self, key):
        if self.gateway_connected(reconnect=False) and key in self._converters:
            self._xsltFactory.removeTransformer(key)
            self._converters.pop(key, None)

    def _parms(self, **kwargs):
        m = self._gateway.jvm.java.util.HashMap()
        for k, v in kwargs.items():
            m[k] = v
        return m

    def transform(self, key, xml, **kwargs):
        """
        Transform the supplied XML using the transform identified by key
        @param key: name of the transform to apply
        @param xml: XML to transform
        @param kwargs: XSLT parameters
        @return: Transform output or None if transform failed
        """
        if key in self._xsltLibrary and self.gateway_connected() and key in self._converters:
            return self._converters[key].transform(xml, self._parms(**kwargs))
        return None



# -*- coding: utf-8 -*-
# Copyright (c) 2014, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the <ORGANIZATION> nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import os
import pyjxslt

xsl1 = """<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    exclude-result-prefixes="xs xd"
    version="2.0">
    <xsl:template match="/doc">
        <xsl:for-each select="entry">
ENTRY: <xsl:value-of select="@id"/>:<xsl:value-of select="."/>
        </xsl:for-each>
    </xsl:template>
</xsl:stylesheet>"""

xsl2 = """<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    exclude-result-prefixes="xs xd"
    version="2.0">
    <xsl:param name="p1">DEVEL</xsl:param>
    <xsl:param name="p2">17</xsl:param>
    <xsl:template match="/doc">
Parm1: <xsl:value-of select="$p1"/>
Parm2: <xsl:value-of select="$p2"/>
        <xsl:for-each select="entry">
ENTRY: <xsl:value-of select="@id"/>:<xsl:value-of select="."/>
        </xsl:for-each>
    </xsl:template>
</xsl:stylesheet>"""

xsl3 = """<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    exclude-result-prefixes="xs xd"
    version="2.0">"""

xsl4 = """<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    exclude-result-prefixes="xs xd"
    version="2.0">JUNK</xsl:stylesheet>"""

xml1 = """<?xml version="1.0" encoding="UTF-8"?>
<doc>
    <entry id='17'>FOO</entry>
    <entry id='42'>BAR</entry>
</doc>"""

xml2 = """<?xml version="1.0" encoding="UTF-8"?>
<doc>
    <entry id='17'>FOO</entry>
    <entry id='42'>BAR</entry>
</docx>"""


class TestGateway(unittest.TestCase):
    gw = pyjxslt.Gateway()

    def testSimple(self):
        self.gw.add_transform('k1', xsl1)
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('k1', xml1))

    def testParms(self):
        self.gw.add_transform('k2', xsl2)
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
Parm1: DEVEL
Parm2: 42
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('k2', xml1, p2=42))
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
Parm1: PROD
Parm2: 42
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('k2', xml1, p2=42, p1="PROD"))

    def testFile(self):
        self.gw.add_transform('k3', os.path.join('data', 'file3.xsl'))
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('k3', xml1))
        self.gw.add_transform('k3', os.path.join('data', 'file3a.xsl'))
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
entry: 17:FOO
entry: 42:BAR""", self.gw.transform('k3', xml1))

    def testReplace(self):
        self.gw.add_transform('k4', xsl1)
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('k4', xml1))
        self.gw.add_transform('k4', xsl2)
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
Parm1: DEVEL
Parm2: 17
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('k4', xml1))

    def testBadXSL(self):
        self.assertRaises(ValueError, self.gw.add_transform, *('e1', xsl3))
        self.assertIsNone(self.gw.add_transform('e2', xsl4))
        self.gw.add_transform('e2', xsl2)
        self.assertEqual("""<?xml version="1.0" encoding="UTF-8"?>
Parm1: DEVEL
Parm2: 17
ENTRY: 17:FOO
ENTRY: 42:BAR""", self.gw.transform('e2', xml1))

    def testBadXML(self):
        self.gw.add_transform('k1', xsl1)
        self.assertTrue(self.gw.transform('k1', xml2)
                        .startswith('ERROR:'))

if __name__ == '__main__':
    unittest.main()




try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# This fails on the actual setup for some unknown reason.  We're going brute force for the moment...
# jarfiles = [f for f in resource_listdir(__name__, 'lib')]
jarfiles = ['lib/py4j-0.9.jar', 'lib/pyjxslt.jar', 'lib/Saxon-HE-9.7.0-1.jar']

setup(
    name='pyjxslt',
    packages=['pyjxslt'],
    package_dir={'pyjxslt': 'src/pyjxslt'},
    package_data={'pyjxslt': ['xsl/*.xsl']},
    version='0.7.0',
    url='http://github.com/CTS2/pyjxslt',
    license='BSD License',
    author='Harold Solbrig',
    author_email='solbrig.harold@mayo.edu',
    description='Python XSLT 2.0 Gateway',
    long_description='Interface package between native python and the Saxon XSLT 2.0 process, which is'
                     ' java based.  This package requires Java 1.7.',
    install_requires=["py4j >=0.9"],
    scripts=['scripts/pyjxslt', 'scripts/testgateway'],
    zip_safe=True,
    include_package_data=True,
    data_files=[('share/pyjxslt', jarfiles)],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
        'Topic :: Text Processing'
    ]
)
